package polyglot.ext.ejava.types;

import java.util.*;

import polyglot.ast.Expr;
import polyglot.ext.ejava.ast.ExtensionApplication;
import polyglot.ext.ejava.ast.ExtensionApplication_c;
import polyglot.ext.jl.ast.CanonicalTypeNode_c;
import polyglot.ext.jl5.types.JL5ConstructorInstance;
import polyglot.ext.jl5.types.JL5Context;
import polyglot.ext.jl5.types.JL5MethodInstance;
import polyglot.ext.jl5.types.JL5TypeSystem_c;
import polyglot.frontend.Source;
import polyglot.main.Report;
import polyglot.types.*;
import polyglot.util.Position;

public class EJavaTypeSystem_c extends JL5TypeSystem_c implements EJavaTypeSystem {

    public ExtensionType findExtension(ClassType from, ClassType to, EJavaContext context,
            Position pos) throws SemanticException {

        if (from instanceof UnextendedType)
            return null;
        List<ExtensionType> extensions = context.extensions();
        ExtensionType r = null;
        if (extensions == null)
            extensions = new ArrayList<ExtensionType>();
        for (ExtensionType et : extensions) {
            if (!isSubtype(from, et.superType()))
                continue;
            for (Iterator ii = et.interfaces().iterator(); ii.hasNext();) {
                ClassType i = (ClassType) ii.next();
                if (isSubtype(i, to)) {
                    if (r != null)
                        throw new SemanticException("ambiguity due to " + "extension application "
                                + "inference", pos);
                    r = et;
                    break;
                }
            }
        }
        if (r != null)
            return r;
        context = (EJavaContext) context.pop();
        if (context == null)
            return null;
        else
            return findExtension(from, to, context, pos);
    }

    public Expr applyExtensions(Expr e, Type to, EJavaContext context) throws SemanticException {

        Type from = e.type();
        if (from instanceof ClassType && to instanceof ClassType) {
            ExtensionType et = findExtension((ClassType) from, (ClassType) to, context, e
                    .position());
            if (et != null
                    && !(e instanceof ExtensionApplication && ((ExtensionApplication) e)
                            .extension().type() == et)) {
                if (isImplicitCastValid(from, to))
                    throw new SemanticException("ambiguity due to extension "
                            + "application inference", e.position());
                e = (Expr) new ExtensionApplication_c(e.position(), e, new CanonicalTypeNode_c(
                        Position.COMPILER_GENERATED, et), false).type(to);
            }
        }
        return e;
    }

    public UnextendedType UnextendedType(ClassType t) {
        return new UnextendedType_c(this, t);
    }

    protected List extensionFriendlyFindAcceptableConstructors(ClassType container, List argTypes,
            EJavaContext context) throws SemanticException {

        // First, try to see if this constructor call would work w/o expanders
        try {
            List r = super.findAcceptableConstructors(container, argTypes, context.currentClass());
            if (r.size() > 0)
                return r;
        } catch (Exception e) {
        }

        // It didn't work - now we try it with expanders

        assert_(container);
        assert_(argTypes);

        List acceptable = new ArrayList();

        if (Report.should_report(Report.types, 2))
            Report.report(2, "Searching type " + container + " for constructor " + container + "("
                    + listToString(argTypes) + ")");

        for (Iterator i = container.constructors().iterator(); i.hasNext();) {
            ConstructorInstance ci = (ConstructorInstance) i.next();

            if (Report.should_report(Report.types, 3))
                Report.report(3, "Trying " + ci);

            if (callValid(ci, argTypes, context) && isAccessible(ci, context.currentClass())) {
                if (Report.should_report(Report.types, 3))
                    Report.report(3, "->acceptable: " + ci);
                acceptable.add(ci);
            }
        }

        return acceptable;
    }

    public ConstructorInstance extensionFriendlyFindConstructor(ClassType container, List argTypes,
            EJavaContext context) throws SemanticException {
        assert_(container);
        assert_(argTypes);
        // System.out.println("fc: "+container+"("+argTypes+") in
        // "+context.currentClass());
        List acceptable;
        while (true) {
            acceptable = extensionFriendlyFindAcceptableConstructors(container, argTypes, context);
            if (acceptable.size() > 0)
                break;
            EJavaContext outerContext = (EJavaContext) context.pop();
            if (outerContext == null)
                break;
            context = outerContext;
        }
        // System.out.println("acceptable: "+acceptable);
        if (acceptable.size() == 0)
            throw new NoMemberException(NoMemberException.CONSTRUCTOR,
                    "No valid constructor found for " + container + "(" + listToString(argTypes)
                            + ").");
        ConstructorInstance ci = (ConstructorInstance) findProcedure(acceptable, container,
                argTypes, context.currentClass());
        if (ci == null)
            throw new NoMemberException(NoMemberException.CONSTRUCTOR, "Reference to " + container
                    + " is ambiguous, multiple " + "constructors match: " + acceptable);
        return ci;
    }

    protected boolean callValid(ProcedureInstance pi, List argTypes, EJavaContext context)
            throws SemanticException {
        boolean ambiguous = false;
        Iterator i1 = pi.formalTypes().iterator(), i2 = argTypes.iterator();
        List extensions = context.extensions();
        if (extensions == null)
            extensions = new LinkedList();
        while (i1.hasNext() && i2.hasNext()) {
            Type t1 = (Type) i1.next();
            Type t2 = (Type) i2.next();
            boolean ok = isImplicitCastValid(t2, t1);
            // Only apply extensions to arguments which haven't been extended
            // explicitly.
            if (!ok && !(t2 instanceof ExtensionType || t2 instanceof UnextendedType))
                for (Iterator i3 = extensions.iterator(); i3.hasNext();) {
                    ExtensionType et = (ExtensionType) i3.next();
                    if (!t2.isSubtype(et.superType()))
                        continue;
                    for (Iterator i4 = et.interfaces().iterator(); i4.hasNext();) {
                        Type t3 = (Type) i4.next();
                        if (isImplicitCastValid(t3, t1)) {
                            if (ok)
                                ambiguous = true;
                            // System.out.println("expander " + et + " has saved
                            // the day for "+pi);
                            ok = true;
                        }
                    }
                }
            if (!ok)
                return false;
        }
        if (!(i1.hasNext() || i2.hasNext())) {
            if (ambiguous)
                throw new SemanticException("ambiguity due to expander application inference");
            else
                return true;
        } else
            return false;
    }

    protected List extensionFriendlyFindAcceptableMethods(ReferenceType container, String name,
            List argTypes, EJavaContext context) throws SemanticException {
        // First, try to see if this method call would work w/o expanders
        try {
            List r = super.findAcceptableMethods(container, name, argTypes, context.currentClass());
            if (r != null && r.size() > 0)
                return r;
        } catch (SemanticException e) {
        }

        // It didn't work... Now we try it with expanders.
        assert_(container);
        assert_(argTypes);

        ClassType currClass = context.currentClass();

        // The list of acceptable methods. These methods are accessible from
        // currClass, the method call is valid, and they are not overridden
        // by an unacceptable method (which can occur with protected methods
        // only).
        List acceptable = new ArrayList();

        // A list of unacceptable methods, where the method call is valid, but
        // the method is not accessible. This list is needed to make sure that
        // the acceptable methods are not overridden by an unacceptable method.
        List unacceptable = new ArrayList();

        Set visitedTypes = new HashSet();

        LinkedList typeQueue = new LinkedList();
        if (currClass instanceof ExtensionType
                && ((ExtensionType) currClass).superType() == container)
            typeQueue.addLast(currClass);
        typeQueue.addLast(container);
        List extensions = context.extensions();
        if (extensions != null) {
            // && !(container instanceof ExtensionType)) {
            // I EXCLUDED THIS LAST PART SO THAT EXTENSIONS CAN USE OTHER
            // EXTENSIONS, but I guess it was wrong!
            // I DON'T KNOW WHY I WROTE THAT ^^^^ , SO I TOOK IT OUT AGAIN, OCT.
            // 27, 2005
            Iterator i = extensions.iterator();
            while (i.hasNext()) {
                ExtensionType et = (ExtensionType) i.next();
                if (container.isSubtype(et.superType()))
                    typeQueue.addLast(et);
            }
        }

        while (!typeQueue.isEmpty()) {
            Type type = (Type) typeQueue.removeFirst();

            if (visitedTypes.contains(type))
                continue;

            visitedTypes.add(type);

            if (Report.should_report(Report.types, 2))
                Report.report(2, "Searching type " + type + " for method " + name + "("
                        + listToString(argTypes) + ")");
            if (!type.isReference())
                throw new SemanticException("Cannot call method in non-reference type " + type
                        + ".");

            for (Iterator i = type.toReference().methods().iterator(); i.hasNext();) {
                MethodInstance mi = (MethodInstance) i.next();

                if (Report.should_report(Report.types, 3))
                    Report.report(3, "Trying " + mi);

                if (methodCallValid(mi, name, argTypes, context)) {
                    if (mi.container() instanceof ExtensionType && container instanceof ClassType
                            && isAccessible(mi, (ClassType) container)
                            || isAccessible(mi, currClass)) {
                        if (Report.should_report(Report.types, 3))
                            Report.report(3, "->acceptable: " + mi + " in " + mi.container());
                        boolean shouldAdd = true;
                        for (Iterator j = acceptable.iterator(); j.hasNext();) {
                            MethodInstance omi = (MethodInstance) j.next();
                            if (mi.name().equals(omi.name())
                                    && mi.formalTypes().equals(omi.formalTypes())) {
                                try {
                                    ExtensionType oet = (ExtensionType) omi.container();
                                    ExtensionType et = (ExtensionType) mi.container();
                                    if (et.extensionName().equals(oet.extensionName())) {
                                        if (et.superType().descendsFrom(oet.superType()))
                                            j.remove();
                                        else {
                                            shouldAdd = false;
                                            break;
                                        }
                                    }
                                } catch (ClassCastException e) {
                                }
                            }
                        }
                        if (shouldAdd) {
                            // System.out.println("adding method " + mi + " from
                            // " + mi.container() + " to acceptable");
                            acceptable.add(mi);
                        }
                    } else {
                        // method call is valid, but the method is
                        // unacceptable.
                        unacceptable.add(mi);
                    }
                }
            }
            if (type.toReference().superType() != null)
                typeQueue.addLast(type.toReference().superType());

            typeQueue.addAll(type.toReference().interfaces());
        }

        // remove any method in acceptable that are overridden by an
        // unacceptable
        // method.
        for (Iterator i = unacceptable.iterator(); i.hasNext();)
            acceptable.removeAll(((MethodInstance) i.next()).overrides());

        /*
         * System.out.println("acceptable: "); for (int i = 0; i <
         * acceptable.size(); i++) { MethodInstance mi =
         * (MethodInstance)acceptable.get(i); System.out.println("
         * "+mi.container()+"'s "+mi); }
         */

        return acceptable;
    }

    /*
     * // invalidate the "bad" findMethod, it's not extension friendly public
     * MethodInstance findMethod(ReferenceType container, String name, List
     * argTypes, ClassType currClass) throws SemanticException { throw new
     * InternalCompilerError("findMethod() should never be called"); }
     */

    public MethodInstance extensionFriendlyFindMethod(ReferenceType container, String name,
            List argTypes, EJavaContext context) throws SemanticException {
        assert_(container);
        assert_(argTypes);
        List acceptable;
        // System.out.println("FM: "+container+"."+name+"("+argTypes+") in
        // "+context.currentClass());
        while (true) {
            // System.out.println("context's extensions:
            // "+((PredJContext)context).extensions());
            acceptable = extensionFriendlyFindAcceptableMethods(container, name, argTypes, context);
            if (acceptable.size() > 0)
                break;
            EJavaContext outerContext = (EJavaContext) context.pop();
            if (outerContext == null)
                break;
            context = outerContext;
        }

        // Remove duplicate abstract procedure instances, because they
        // are not really ambiguities - there's only one implementation ever.
        // System.out.println("acceptable0: "+acceptable);
        int i = 0;
        while (i < acceptable.size()) {
            ProcedureInstance pi = (ProcedureInstance) acceptable.get(i);
            if (pi.flags().isAbstract()) {
                int j = i + 1;
                while (j < acceptable.size()) {
                    ProcedureInstance pj = (ProcedureInstance) acceptable.get(j);
                    if (!pj.flags().isAbstract() || pi == pj)
                        break;
                    if (pi.signature().equals(pj.signature()))
                        acceptable.remove(pj);
                    j += 1;
                }
            }
            i += 1;
        }
        // System.out.println("acceptable1: "+acceptable);

        if (acceptable.size() == 0)
            throw new NoMemberException(NoMemberException.METHOD, "No valid method call found for "
                    + name + "(" + listToString(argTypes) + ") in " + container + ".");
        MethodInstance mi = (MethodInstance) findProcedure(acceptable, container, argTypes, context
                .currentClass());
        if (mi == null)
            throw new SemanticException("Reference to " + name
                    + " is ambiguous, multiple methods match: " + acceptable);
        // System.out.println("chosen: "+mi.container()+(mi.container()
        // instanceof ExtensionType ? "(extension)" : "")+"'s "+mi);
        return mi;
    }

    protected boolean methodCallValid(MethodInstance mi, String name, List argTypes,
            EJavaContext context) throws SemanticException {
        return name.equals(mi.name()) && callValid(mi, argTypes, context);
    }

    public ExtensionType createExtensionType(LazyClassInitializer init, Source fromSource) {
        return new ExtensionType_c(this, init, fromSource);
    }

    @Override
    public Context createContext() {
        return new EJavaContext_c(this);
    }

    protected EJavaContext context;

    public EJavaContext getCurrentContext() {
        return context;
    }

    public void setCurrentContext(EJavaContext ctx) {
        context = ctx;
    }


    protected boolean useExpanders = false;
    public boolean getUseExpanders() {
        return useExpanders;
    }

    public void setUseExpanders(boolean useExpanders) {
        this.useExpanders = useExpanders;
    }

    @Override
    protected Set<JL5MethodInstance> findMethodsNamed(ReferenceType container, String name) {
        Set<JL5MethodInstance> res = super.findMethodsNamed(container, name);
        if (res.size() == 0 && getUseExpanders()) {
            setUseExpanders(false);
            for (ExtensionType et : getCurrentContext().extensions()) {
                if (et.implicitlyApplicable(container)) {
                        res.addAll(findMethodsNamed(et, name));
                }
            }
            setUseExpanders(true);
        }
        return res;
    }

    @Override
    public boolean isSubtype(Type t1, Type t2) {
        //implement implicit expander expansion
        boolean oldExp = getUseExpanders();
        setUseExpanders(false);
            boolean r = super.isSubtype(t1, t2);
        setUseExpanders(oldExp);
        if (!r && getUseExpanders()) {
            //try expanders.
            for (ExtensionType et : getCurrentContext().extensions()) {
                if (et.implicitlyApplicable(t1)) {
                    setUseExpanders(false);
                    r = isSubtype(et, t2);
                    setUseExpanders(true);
                    if (r) return true;
                }
            }
        }
        return r;
    }

    @Override
    public JL5ConstructorInstance findJL5Constructor(ClassType ct, List<Type> paramTypes, List<Type> explicitTypeArgs, JL5Context context) throws SemanticException {
        //first try without expanders
        setUseExpanders(false);
        try {
            return super.findJL5Constructor(ct, paramTypes, explicitTypeArgs, context);
        } catch (SemanticException e) {
        }
        //now try with expanders
        setUseExpanders(true);
        try {
            return super.findJL5Constructor(ct, paramTypes, explicitTypeArgs, context);
        } finally {
            setUseExpanders(false);
        }
    }

    @Override
    public JL5MethodInstance findJL5Method(ReferenceType container, String name, List<Type> paramTypes, List<Type> explicitTypeArgTypes, JL5Context context) throws SemanticException {
        //try without expanders first
        setUseExpanders(false);
        try {
            return super.findJL5Method(container, name, paramTypes, explicitTypeArgTypes, context);
        } catch (SemanticException e) {
        }
        //now try with expanders. this will affect findMethodsNamed();
        setUseExpanders(true);
        try {
            return super.findJL5Method(container, name, paramTypes, explicitTypeArgTypes, context);
        } finally {
            setUseExpanders(false);
        }
    }

    @Override
    protected Set<FieldInstance> findFields(ReferenceType container, java.lang.String name) {
        Set<FieldInstance> res = super.findFields(container, name);
        if (res.size() == 0 && getUseExpanders()) {
            //now try expanders
            setUseExpanders(false);
            for (ExtensionType et : getCurrentContext().extensions()) {
                if (et.implicitlyApplicable(container)) {
                    res.addAll(findFields(et, name));
                }
            }
            setUseExpanders(true);
        }
        return res;
    }

    @Override
    public FieldInstance findFieldOrEnum(ReferenceType container, String name, ClassType currClass) throws SemanticException {
        //try without expanders
        setUseExpanders(false);
        try {
            return super.findFieldOrEnum(container, name, currClass);
        } catch (SemanticException e) {
        }
        //now try with expanders. this will affect findFields
        setUseExpanders(true);
        try {
            return super.findFieldOrEnum(container, name, currClass);
        } finally {
            setUseExpanders(false);
        }
    }
    
}
